!-----------------------------------------------------------------------!
!                                                                       !
!                                                                       !
!                      Algorithmes de descente                          !
!                                                                       !
!                                                                       !
!-----------------------------------------------------------------------!

module M_Descente
  use M_func
  use M_types
  implicit none
contains
  subroutine Uzawa_penalisation(Data)
    type(Donnees),intent(in)       ::Data
    real,dimension(data%ndim)      ::xk
    real                           ::f
    integer                        ::Nf,k,i
    real                           ::epsilon

    !determination de Nf

    select case(Data%Nmeth)
    case(0)
       !optimisation sans contrainte
       Nf=Data%Nfunc
       call Descente(Nf,xk,f,Data)
    case(1)
       !methode de penalisation
       Nf=10
       epsilon=0.3
       do k=1,30
          call Init_Penalisation(epsilon)
          call Descente(Nf,xk,f,Data)
          print*,'epsilon=',epsilon
          epsilon=2.*epsilon

          print*,"point en lequel est atteint le minimum"
          print*,xk(:)
          print*,"valeur du minimum"
          call func(xk,f,Nf,Data)
          print*,f
          print*,'valeur de la fonction'
          call func(xk,f,Data%Nfunc,Data)
          print*,f
          print*,'valeur de la contrainte'
          do i=1,data%cont
             print*,'Valeur de la contrainte',i
             print*,'data%contnum(i)',data%contnum(i)
             call func(xk,f,Data%contnum(i),Data)
             print*,f
          enddo

          print*,'appuer sur entree pour lancer une nouvelle methode de &
               & penalisation'
          read(*,*)
       enddo
    case(2)
       !methode d'Uzawa
       Nf=20
    case default
       print*,"la methode n'est pas reconnue dans Uzawa_Penalisation"
       print*,'fin du programme'
       stop
    end select    
    
  end subroutine Uzawa_penalisation

  subroutine Descente(Nf,xk,f,Data)
    type(Donnees),intent(in)       ::Data
    !variable locale
    real,dimension(data%ndim)      ::xk,dk,gradient
    real                           ::f,normgrad,alpha
    integer                        ::k,j
    integer                        ::Nf

    !point initial

    xk(:)=1.

    call Gradientf(xk,gradient,Nf,Data)
    normgrad=sqrt(sum(gradient(:)*gradient(:)))

    k=0


    do while((k.le.Data%Nmax).and.(normgrad**2.ge.Data%tol))
       
       !choix de la direction de descente
       call Gradientf(xk,dk,Nf,Data)
       
       !recherche du pas de descente
       print*,'avant goldstein'
       call Goldstein(xk,alpha,dk,Nf,Data)
       print*,'apres goldstein'

       !iteration
       xk(:)=xk(:)-alpha*dk(:)
       print*,'||x_{k+1}-x_k||',alpha*sqrt(sum(dk(:)*dk(:)))
       !estimation de l'erreur
       call Gradientf(xk,gradient,Nf,data)
       normgrad=sqrt(sum(gradient(:)*gradient(:)))
       k=k+1
       print*,"k=",k
       call func(xk,f,Nf,Data)
       !ecriture d'une sortie

       print*,'xk(:)',xk(:)
       print*,'f(xk)',f
       print*,"normgrad",normgrad
       print*,'gradient',gradient(:)
       print*,'dk(:)',dk(:)
    enddo


  end subroutine Descente

!-----------------------------------------------------------------------!
!                                                                       !
!                                                                       !
!                      Recherche lineaire de pas                        !
!                                                                       !
!                                                                       !
!-----------------------------------------------------------------------!

!****************************
!Methode de Goldstein
!****************************

  subroutine Goldstein(x,alpha,dk,Nf,Data)
    type(Donnees)               ::Data
    real,dimension(data%ndim)   ::x,dk
    real                        ::alpha
    integer                     ::Nf
    !variables locales
    real                        ::alphamin,alphamax,rho
    real,dimension(data%ndim)   ::gradient,xalpha
    real                        ::falpha
    real                        ::etamoins,etaplus,f,fmax

    !lecture du parametre
    rho=0.2

    !on a besoin du gradient et de la direction de descente
    etaplus =-rho*(sum(dk(:)*dk(:)))
    etamoins=-(1-rho)*(sum(dk(:)*dk(:)))

!!$    print*,'etaplus',etaplus
!!$    print*,'etamoins',etamoins

    if (etaplus.le.etamoins) then
       print*,'probleme dans Goldstein'
       print*,'etaplus',etaplus
       print*,'etamoins',etamoins
       stop
    endif
    alphamin=0.


    call func(x,f,Nf,data)


    !recherche d'un alphamax
    alphamax=Data%pas
    xalpha=x(:)-alphamax*dk(:)
    call func(xalpha,fmax,Nf,data)


    do while ((fmax-f).lt.(alphamax*etaplus))
       alphamax=2.*alphamax
       xalpha=x(:)-alphamax*dk(:)
       call func(xalpha,fmax,Nf,data)
    enddo

!!$    print*,'alphamax=',alphamax
!!$    print*,'falphamax',fmax

    !recherche d'un alpha correct
    alpha=(alphamax+alphamin)/2.
    xalpha=x(:)-alpha*dk(:)
    call func(xalpha,falpha,Nf,data)

!!$    print*,'falphamin=',falpha
!!$    
!!$    print*,'f=',f

    do while (((falpha-f).le.(alpha*etamoins)).or.((falpha-f).ge.(alpha*etaplus)))

!!$       print*,'alphamin',alphamin
!!$       print*,'alphamax',alphamax
!!$       print*,'falpha',falpha
!!$       print*,'f+alpha*etamoins',f+alpha*etamoins
!!$       print*,'f+alpha*etaplus',f+alpha*etaplus

       
       if (falpha.le.(f+alpha*etamoins)) then
          alphamin=alpha
       else
          alphamax=alpha
       endif
       alpha=(alphamax+alphamin)/2.
       xalpha=x(:)-alpha*dk(:)
!!$       print*,'x(:)',x(:)
!!$       print*,'xalpha(:)',xalpha(:)
       call func(xalpha,falpha,Nf,data)

       if (alpha.le.epsilon(1.)) then
          print*,'arret dans la methode de Goldstein'
          print*,'alphamin',alphamin
          print*,'alphamax',alphamax
          print*,'Cette erreur peut etre due a une mauvaise approximation &
               & du gradient'
          print*,'valeur de la fonction'
          print*,f
          print*,'etaplus =',etaplus
          print*,'etamoins =',etamoins
          stop
       endif

    enddo

    alpha=(alphamin+alphamax)/2.

  end subroutine Goldstein


!-----------------------------------------------------------------------!
!                                                                       !
!                                                                       !
!                      Calcul du gradient de f                          !
!                                                                       !
!                                                                       !
!-----------------------------------------------------------------------!

  subroutine Gradientf(point,gradient,Nf,Data)
    type(Donnees),intent(in)    ::Data
    real,dimension(Data%Ndim)   ::point,gradient
    real                        ::epsilon
    integer                     ::Nf
    !variables locales
    integer                     ::k
    real,dimension(Data%Ndim)   ::eps,pointplus
    real                        ::f,fplus

    epsilon=Data%eps
    gradient(:)=0.

    do k=1,Data%Ndim
       eps(:)=0.
       eps(k)=epsilon
       call func(point,f,Nf,data)
       pointplus(:)=point(:)+eps(:)
       call func(pointplus,fplus,Nf,data)
!!$       if (k.eq.1) then
!!$          print*,'fplus=',fplus
!!$          print*,'f=',f
!!$          print*,'fplus-f',fplus-f
!!$          print*,'(fplus-f)/epsilon',(fplus-f)/epsilon
!!$       endif
       gradient(k)=(fplus-f)/epsilon
    enddo
    
  end subroutine Gradientf
end module M_Descente
