module M_func
  use M_types
  implicit none
  real,private                :: epsilonfixe
  real,dimension(:),pointer   :: lambda,mu
  
contains
  subroutine Dimfunc(Data)
    type(Donnees)             ::Data
    
    select case (Data%Nfunc)
    case(1,2,3,4)
       Data%Ndim=2
    case(6:8)
       Data%Ndim=3
    case default
       print*,"le numero de fonction est inconnu dans Dimfunc"
       stop
    end select
    
  end subroutine Dimfunc
  
  subroutine func(entree,f,Nf,Data)
    type(Donnees),intent(in)  ::Data
    real,dimension(Data%Ndim) ::entree
    integer                   ::Nf
    real                      ::f
    
    select case(Nf)
    case(1:8)
       call funcs(entree,f,Nf,Data)
    case(10)
       !la fonction de penalisation, avec epsilon fixe
       !print*,'fonction de penalisation avec epsilon fixe'
       !print*,'non programm'
       !stop
       call Penalisation(entree,epsilonfixe,f,Data)
    case(20)
       !le lagrangien avec mu et lambda fixes
       print*,'Lagrangien avec lambda et mu fixs'
       print*,'non programm'
       stop
    case default
       print*,"le numero de fonction est inconnu dans func"
       print*,'Data%Nfunc=',Data%Nfunc
       print*,'Nf=',Nf
       stop
    end select
    
    
  end subroutine func
  
  subroutine funcs(entree,f,Nf,Data)
    type(Donnees),intent(in)  ::Data
    real,dimension(Data%Ndim) ::entree
    real                      ::f
    integer                   ::Nf
    !variables locales
    real                      ::x,y,z
    real                      ::x1,y1,z1
    

    x=entree(1)
    y=entree(2)

    select case(Nf)
    case(1)
       f=x**2+y**2
!       f=x**2-5*x*y+y**4-25*x-8*y
    case(2)
       f=5*x**2+5*y**2-x*y-11*x+11*y+11
    case(3)
       f=(x**4-3)**2+y**4
    case(4)
       f=x**4-4*x**3+6*(x**2+y**2)-4*(x+y)
    case(5)
       f=x+y-1
    case(6)
       z=entree(3)
       x1=0.01*x+0.0036*y+0.0015*z 
       y1=0.0036*x+0.0144*y+0.0072*z
       z1=0.0015*x+0.0072*y + 0.0225*z
       f=x*x1+y*y1+z*z1
    case(7)
       z=entree(3)
       f=0.05*x+0.08*y+0.10*z-Data%r
    case(8)
       z=entree(3)
       f=x+y+z-1
    case default
       print*,"le numero de fonction est inconnu dans func"
       print*,'Data%Nfunc=',Data%Nfunc
       print*,'Nf=',Nf
       stop
    end select

  end subroutine funcs

  subroutine Lagrangien(x,lambda,mu,f,Data)
    type(Donnees),intent(in)  ::Data
    real,dimension(Data%Ndim)   ::x
    real,dimension(Data%cont)   ::lambda
    real,dimension(Data%Ndim)   ::mu
    real                        ::f,g

    integer                     ::i,k

    f=0.

    !partie 'fonctionnelle'
    call func(x,g,Data%Nfunc,Data)

    f=f+g

    !partie 'contraintes egalite'
    do k=1,Data%cont
       i=Data%contnum(k)
       call func(x,g,i,Data)
       f=f+lambda(k)*g
    enddo

    if (Data%pos) then
       !ajouter les contraintes inegalites sur les coefficients
       !dans le Lagrangien
       do k=1,Data%Ndim
          f=f+mu(k)*(-x(k))
       enddo
    endif

  end subroutine Lagrangien

  subroutine Penalisation(x,epsilon,f,Data)
    type(donnees),intent(in)       ::Data
    real,dimension(Data%Ndim)      ::x
    real                           ::epsilon,f
    
    real                           ::g
    integer                        ::i,k

    f=0.

    !partie 'fonctionnelle'
    call func(x,g,Data%Nfunc,Data)
    
    f=f+g
    
    !partie 'contraintes egalite'
    do k=1,Data%cont
       i=Data%contnum(k)
       call func(x,g,i,Data)
       f=f+epsilon*(g**2)
    enddo
  end subroutine Penalisation

  subroutine Init_Penalisation(epsilon)
    real                ::epsilon
    epsilonfixe=epsilon

  end subroutine Init_Penalisation

end module M_func
